#include "ogreapplication.h"
#include "lightwibbler.h"
#include <OgreErrorDialog.h>

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Interface of MyApplication
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class MyApplication : public OgreApplication
{
public:
  MyApplication();
  virtual void init();
};

using namespace Ogre;
using namespace std;

//************************************************************
// Implementation of MyApplication
//************************************************************
MyApplication::MyApplication()
  : OgreApplication("My Application")
{
}
void
MyApplication::init()
{
  mSceneManager->setAmbientLight(ColourValue(0.7f,0.7f,1.0f));
  
  // Adds a robot      
  Entity *robot = mSceneManager->createEntity("Robot","robot.mesh");
  SceneNode* robotNode = mSceneManager->getRootSceneNode()->createChildSceneNode("RobotNode");
  robotNode->attachObject(robot);
  Real r = robot->getBoundingRadius();
  
  // Adds a ground  
  MeshManager::getSingleton().createPlane("plane.mesh","Custom",Plane(Vector3::UNIT_Z,Vector3::ZERO),2*r,2*r,20,20);
  Entity* ground = mSceneManager->createEntity("Ground","plane.mesh");
  ground->setMaterialName("Examples/Rocky");
  SceneNode* groundNode = mSceneManager->getRootSceneNode()->createChildSceneNode();
  groundNode->attachObject(ground);
  groundNode->setOrientation(Quaternion(Radian(-Math::HALF_PI),Vector3::UNIT_X));
  
  // Adds a light 
  Light* light = mSceneManager->createLight("Light1"); 
  light->setType(Light::LT_SPOTLIGHT);
  light->setDiffuseColour(ColourValue(0.25f,0.25f,0.0f));
  light->setSpecularColour(ColourValue(0.25f,0.25f,0.0f));
  light->setAttenuation(8000,1,0.0005,0);
  light->setSpotlightRange(Ogre::Degree(60), Ogre::Degree(70));
  light->setDirection(Vector3::NEGATIVE_UNIT_Y);

  SceneNode* lightNode = robotNode->createChildSceneNode("LightNode");
  lightNode->attachObject(light);
  Real y = 1.2f*r;
  lightNode->setPosition(0,y,0);

  BillboardSet* lightbillboardset = mSceneManager->createBillboardSet("lightbbs", 1);
  lightbillboardset->setMaterialName("Examples/Flare");
  Billboard* lightbillboard = lightbillboardset->createBillboard(0,0,0,ColourValue(0.5,0.3,0.0f));
  lightNode->attachObject(lightbillboardset);
  
  // Activate shadows
  mSceneManager->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_MODULATIVE);
  mSceneManager->setShadowColour(Ogre::ColourValue(0.5f,0.5f,0.5f));  
  ground->setCastShadows(false);
  robot->setCastShadows(true);

  // Animating the light
  Animation* animation = mSceneManager->createAnimation("LightAnim",10.0f);
  animation->setInterpolationMode(Animation::IM_SPLINE);
  NodeAnimationTrack* track = animation->createNodeTrack(0,lightNode);

  Real x = 0.6f*r;
  Real z = 0.6f*r;
  TransformKeyFrame* key;
  key = track->createNodeKeyFrame(0.0f);
  key->setTranslate(Vector3(-x, y,-z));
  key = track->createNodeKeyFrame(2.5f);
  key->setTranslate(Vector3( x, y,-z));
  key = track->createNodeKeyFrame(5.0f);
  key ->setTranslate(Vector3( x, y, z));
  key = track->createNodeKeyFrame(7.5f);
  key->setTranslate(Vector3(-x, y, z));
  key = track->createNodeKeyFrame(10.0f);
  key->setTranslate(Vector3(-x, y,-z));
  manage(mSceneManager->createAnimationState("LightAnim"),true,true);
  lightNode->setAutoTracking(true,robotNode,Vector3::NEGATIVE_UNIT_Y);

  // Animate the light color
  {
    ControllerFunctionRealPtr func = ControllerFunctionRealPtr(new WaveformControllerFunction(Ogre::WFT_SINE,0.75,0.5));
    ControllerValueRealPtr val     = ControllerValueRealPtr(new LightWibbler(light,lightbillboard,
                                                                             ColourValue(0.3f,0.0f,0.0f),
                                                                             ColourValue(0.5f,0.3f,0.1f)));
    ControllerManager& m = ControllerManager::getSingleton();
    Controller<Real>* controller   = m.createController(m.getFrameTimeSource(),val,func);
  }
  // Animate the robot
  manage(robot->getAnimationState("Idle"),true,true);

  // Add a skybox
  mSceneManager->setSkyBox(true,"Examples/StormySkyBox",50);
}

int main(int, char **)
{
  MyApplication application;
  try
  {
    application.addResourceLocationFrom("../../resources.cfg");
    application.create("../../plugins.cfg","../ogre.cfg","ogre.log",ST_GENERIC);
    application.run();
  }
  catch (Exception& e)
  {
    ErrorDialog* dlg = PlatformManager::getSingleton().createErrorDialog();
    dlg->display(e.getFullDescription());
    delete dlg;
    return 1;
  }
  return 0;
}

